package com.terraformersmc.terrestria.feature.trees;

import com.mojang.datafixers.Dynamic;
import com.terraformersmc.terrestria.feature.TreeDefinition;
import com.terraformersmc.terrestria.init.TerrestriaBlocks;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2944;
import net.minecraft.class_3111;
import net.minecraft.class_3341;
import net.minecraft.class_3481;
import net.minecraft.class_3746;
import net.minecraft.class_3747;

public class PalmTreeFeature extends class_2944<class_3111> {
	private TreeDefinition.WithBark tree;

	public PalmTreeFeature(Function<Dynamic<?>, ? extends class_3111> function, boolean notify, TreeDefinition.WithBark tree) {
		super(function, notify);

		this.tree = tree;
	}

	private static class_2350 spiral(class_2350 direction, boolean invert) {
		switch (direction) {
			case field_11034:
				return invert ? class_2350.field_11043 : class_2350.field_11035;
			case field_11039:
				return invert ? class_2350.field_11035 : class_2350.field_11043;
			case field_11043:
				return invert ? class_2350.field_11039 : class_2350.field_11034;
			case field_11035:
			default:
				return invert ? class_2350.field_11034 : class_2350.field_11039;
		}
	}

	public PalmTreeFeature sapling() {
		return new PalmTreeFeature(class_3111::method_13565, true, tree);
	}

	@Override
	public boolean method_12775(Set<class_2338> blocks, class_3747 world, Random rand, class_2338 origin, class_3341 boundingBox) {
		// Total trunk height
		int height = rand.nextInt(5) + 8;

		if (origin.method_10264() + height + 1 > 256 || origin.method_10264() < 1) {
			return false;
		}

		class_2338 below = origin.method_10074();

		boolean sand = false;

		if (!method_16430(world, below)) {
			if(world.method_16358(below, state -> state.method_11602(class_3481.field_15466))) {
				sand = true;
			} else {
				return false;
			}
		}

		if(!check(world, origin, height)) {
			return false;
		}

		if(!sand) {
			method_16427(world, below);
		}

		class_2338.class_2339 pos = new class_2338.class_2339(origin);
		growTrunk(blocks, world, pos, height, rand, boundingBox);
		growLeaves(blocks, world, pos, rand, boundingBox);

		return true;
	}

	private boolean check(class_3746 world, class_2338 origin, int height) {
		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for(int y = 0; y < height; y++) {
			int radius = y >= 2 ? 1 : 0;

			for(int z = -radius; z <= radius; z++) {
				for(int x = -radius; x <= radius; x++) {
					pos.method_10101(origin).method_10100(x, y, z);

					if(!method_16432(world, pos)) {
						return false;
					}
				}
			}
		}

		return true;
	}

	// Grows the bent trunk of the tree.
	private void growTrunk(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, Random rand, class_3341 boundingBox) {
		for (int i = 0; i < 2; i++) {
			method_12773(blocks, world, pos, tree.getLog(), boundingBox);
			pos.method_10098(class_2350.field_11036);
		}

		int run = 0;

		int velocityX = rand.nextBoolean() ? 1 : -1;
		int velocityZ = rand.nextBoolean() ? 1 : -1;

		for (int i = 2; i < height; i++) {
			if(!method_16432(world, pos)) {
				return;
			}

			if (run++ == 3) {
				method_12773(blocks, world, pos, tree.getBark(), boundingBox);

				if (rand.nextBoolean()) {
					pos.method_10100(velocityX, 0, 0);
				} else {
					pos.method_10100(0, 0, velocityZ);
				}
				run = 0;

				if(!method_16432(world, pos)) {
					return;
				}

				method_12773(blocks, world, pos, tree.getBark(), boundingBox);
			} else {
				method_12773(blocks, world, pos, tree.getLog(), boundingBox);
			}

			pos.method_10098(class_2350.field_11036);
		}
	}

	private void tryPlaceLeaves(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, class_3341 boundingBox) {
		if (class_2944.method_16420(world, pos)) {
			method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
		}
	}

	private void growLeaves(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, Random rand, class_3341 boundingBox) {
		class_2338 center = pos.method_10062();

		if(method_16432(world, pos)) {
			method_12773(blocks, world, pos, tree.getLog(), boundingBox);
		}

		tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(0, 1, 0), boundingBox);
		tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(1, 1, 0), boundingBox);
		tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(0, 1, 1), boundingBox);
		tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(-1, 1, 0), boundingBox);
		tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(0, 1, -1), boundingBox);

		boolean invertLeafSpiral = rand.nextBoolean();

		for (int dZ = -1; dZ < 2; dZ++) {
			for (int dX = -1; dX < 2; dX++) {
				tryPlaceLeaves(blocks, world, pos.method_10101(center).method_10100(dZ, 0, dX), boundingBox);
			}
		}

		for (int d = 0; d < 4; d++) {
			class_2350 direction = class_2350.method_10139(d);

			pos.method_10101(center).method_10104(direction, 2);
			placeSpiral(blocks, world, pos, direction, !invertLeafSpiral, boundingBox);

			pos.method_10101(center).method_10104(direction, 3);
			placeSpiral(blocks, world, pos, direction, invertLeafSpiral, boundingBox);
		}
	}

	private void placeSpiral(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, class_2350 direction, boolean invertLeafSpiral, class_3341 boundingBox) {
		tryPlaceLeaves(blocks, world, pos, boundingBox);

		class_2350 spiral = spiral(direction, invertLeafSpiral);
		tryPlaceLeaves(blocks, world, pos.method_10098(spiral), boundingBox);

		for (int i = 0; i < 2; i++) {
			tryPlaceLeaves(blocks, world, pos.method_10098(class_2350.field_11033), boundingBox);
		}
	}
}
