package com.terraformersmc.terrestria.feature.trees.templates;

import com.mojang.datafixers.Dynamic;
import com.terraformersmc.terraform.block.QuarterLogBlock;
import com.terraformersmc.terraform.util.Shapes;
import com.terraformersmc.terrestria.feature.TreeDefinition;
import com.terraformersmc.terrestria.feature.trees.components.Roots;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2476;
import net.minecraft.class_2525;
import net.minecraft.class_2944;
import net.minecraft.class_3111;
import net.minecraft.class_3341;
import net.minecraft.class_3486;
import net.minecraft.class_3532;
import net.minecraft.class_3746;
import net.minecraft.class_3747;

public class CanopyTreeFeatureMega extends class_2944<class_3111> implements Roots {
	private TreeDefinition.Mega tree;

	public CanopyTreeFeatureMega(Function<Dynamic<?>, ? extends class_3111> function, boolean notify, TreeDefinition.Mega tree) {
		super(function, notify);

		this.tree = tree;
	}

	protected static boolean method_16432(class_3746 world, class_2338 pos) {
		return class_2944.method_16432(world, pos) || world.method_16358(pos,
				state -> state.method_11618().method_15772().method_15791(class_3486.field_15517) || state.method_11614() instanceof class_2476
		);
	}

	public CanopyTreeFeatureMega sapling() {
		return new CanopyTreeFeatureMega(class_3111::method_13565, true, tree);
	}

	@Override
	public boolean method_12775(Set<class_2338> blocks, class_3747 world, Random rand, class_2338 origin, class_3341 boundingBox) {
		// Total trunk height
		int height = getHeight(rand);

		// How much "bare trunk" there will be. (2-3)
		int bareTrunkHeight = getBareTrunkHeight(rand);

		if (origin.method_10264() + height + 1 > 256 || origin.method_10264() < 1) {
			return false;
		}

		class_2338 below = origin.method_10074();

		int down = 0;
		while (world.method_16358(below, state -> state.method_11618().method_15772().method_15791(class_3486.field_15517) || state.method_11614() instanceof class_2476)) {
			below = below.method_10074();
			down++;
		}

		height += down * 3 / 2;
		bareTrunkHeight += down * 3 / 2;

		origin = below.method_10084();

		for (int dZ = 0; dZ < 2; dZ++) {
			for (int dX = 0; dX < 2; dX++) {
				below = origin.method_10069(dX, -1, dZ);

				if (!method_16430(world, below)) {
					return false;
				}
			}
		}

		if (!checkForObstructions(world, origin, height, bareTrunkHeight)) {
			return false;
		}

		for (int dZ = 0; dZ < 2; dZ++) {
			for (int dX = 0; dX < 2; dX++) {
				method_16427(world, origin.method_10069(dX, -1, dZ));
			}
		}

		growTrunk(blocks, world, new class_2338.class_2339(origin), height / 2, boundingBox);

		class_2338.class_2339 pos = new class_2338.class_2339(origin).method_10104(class_2350.field_11036, bareTrunkHeight);
		growBranches(blocks, world, pos, height - bareTrunkHeight, height / 2 - bareTrunkHeight, rand, boundingBox);

		pos.method_10101(origin).method_10104(class_2350.field_11033, 2);
		growRoots(blocks, world, pos, bareTrunkHeight + 2, rand, boundingBox);

		return true;
	}

	private boolean checkForObstructions(class_3746 world, class_2338 origin, int height, int bareTrunkHeight) {
		class_2338.class_2339 pos = new class_2338.class_2339(origin);

		for (int dY = 0; dY < bareTrunkHeight; dY++) {
			for (int dZ = 0; dZ < 2; dZ++) {
				for (int dX = 0; dX < 2; dX++) {

					pos.method_10103(origin.method_10263() + dX, origin.method_10264() + dY, origin.method_10260() + dZ);

					if (!method_16432(world, pos)) {
						return false;
					}
				}
			}
		}

		for (int dY = bareTrunkHeight; dY < height + 1; dY++) {
			for (int dZ = -7; dZ < 8; dZ++) {
				for (int dX = -7; dX < 8; dX++) {
					pos.method_10103(origin.method_10263() + dX, origin.method_10264() + dY, origin.method_10260() + dZ);

					if (!method_16432(world, pos)) {
						return false;
					}
				}
			}
		}

		return true;
	}

	private void growTrunk(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, class_3341 boundingBox) {
		int x = pos.method_10263();
		int y = pos.method_10264();
		int z = pos.method_10260();

		for (int i = 0; i < height; i++) {
			pos.method_10103(x, y + i, z);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.NORTHWEST), boundingBox);

			pos.method_10103(x + 1, y + i, z);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.NORTHEAST), boundingBox);

			pos.method_10103(x, y + i, z + 1);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.SOUTHWEST), boundingBox);

			pos.method_10103(x + 1, y + i, z + 1);
			method_12773(blocks, world, pos, tree.getLogQuarter().method_11657(QuarterLogBlock.BARK_SIDE, QuarterLogBlock.BarkSide.SOUTHEAST), boundingBox);
		}
	}

	private void growBranches(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int height, int trunkHeight, Random rand, class_3341 boundingBox) {
		int branches = rand.nextInt(10) + 15;

		int[] offset = new int[3];
		class_2338 origin = pos.method_10062();

		for (int branch = 0; branch < branches; branch++) {
			int startOffsetY = rand.nextInt(height);

			offset[Shapes.X] = class_3532.method_15340(rand.nextInt(13) - 6, -5, 5);
			offset[Shapes.Y] = rand.nextInt(height - 3) + trunkHeight / 2;
			offset[Shapes.Z] = class_3532.method_15340(rand.nextInt(13) - 6, -5, 5);

			int alignX = offset[Shapes.X] > 0 ? 1 : 0;
			int alignZ = offset[Shapes.Z] > 0 ? 1 : 0;

			pos.method_10101(origin);
			pos.method_10069(alignX, startOffsetY, alignZ);

			Shapes.line(pos, offset, position -> {
				if (!world.method_16358(position, candidate -> candidate.method_11614() instanceof QuarterLogBlock)) {
					method_12773(blocks, world, pos, tree.getLog(), boundingBox);
				}
			});

			int maxRadius = Math.min(
					Math.min(
							Math.min(8 - offset[Shapes.X], 8 - offset[Shapes.Z]),
							Math.min(8 + offset[Shapes.X] - 1, 8 + offset[Shapes.Z] - 1)
					),
					4
			);

			int x = pos.method_10263();
			int z = pos.method_10260();
			int radius = maxRadius == 4 ? 3 + rand.nextInt(2) : 2 + rand.nextInt(maxRadius - 1);

			pos.method_10098(class_2350.field_11033);

			for (int i = -1; i <= 1; i++) {
				pos.method_10103(x, pos.method_10264(), z);

				int layerRadius = (int) Math.floor(radius * Math.cos(i * Math.PI / 6));

				Shapes.circle(pos, layerRadius, position -> {
					if (class_2944.method_16420(world, pos)) {
						method_12773(blocks, world, pos, tree.getLeaves(), boundingBox);
					}
				});

				pos.method_10098(class_2350.field_11036);
			}
		}
	}

	public void growRoots(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 pos, int baseTrunkHeight, Random rand, class_3341 boundingBox) {
		int x = pos.method_10263();
		int y = pos.method_10264();
		int z = pos.method_10260();

		tryGrowRoot(blocks, world, pos.method_10103(x - 1, y, z + rand.nextInt(2)), baseTrunkHeight, rand, boundingBox);
		tryGrowRoot(blocks, world, pos.method_10103(x + 2, y, z + rand.nextInt(2)), baseTrunkHeight, rand, boundingBox);
		tryGrowRoot(blocks, world, pos.method_10103(x + rand.nextInt(2), y, z - 1), baseTrunkHeight, rand, boundingBox);
		tryGrowRoot(blocks, world, pos.method_10103(x + rand.nextInt(2), y, z + 2), baseTrunkHeight, rand, boundingBox);
	}

	public void tryGrowRoot(Set<class_2338> blocks, class_3747 world, class_2338.class_2339 bottom, int baseTrunkHeight, Random rand, class_3341 boundingBox) {
		if (rand.nextInt(5) == 0) {
			return;
		}

		int height = baseTrunkHeight + rand.nextInt(2) - 3;

		for (int i = 0; i < height; i++) {
			if (method_16432(world, bottom) || class_2944.method_16425(world, bottom) || world.method_16358(bottom, state -> state.method_11614() instanceof class_2525)) {
				method_12773(blocks, world, bottom, tree.getBark(), boundingBox);
			}

			bottom.method_10098(class_2350.field_11036);
		}
	}

	public int getHeight(Random rand) {
		return rand.nextInt(8) + 19;
	}

	public int getBareTrunkHeight(Random rand) {
		return 2 + rand.nextInt(2);
	}
}
