package com.terraformersmc.terrestria.feature.volcano;

import com.terraformersmc.terrestria.init.TerrestriaBiomes;
import com.terraformersmc.terrestria.init.TerrestriaBlocks;
import com.terraformersmc.terraform.noise.OpenSimplexNoise;
import com.terraformersmc.terraform.noise.SimpleRadialNoise;
import com.terraformersmc.terrestria.init.TerrestriaFeatures;
import net.minecraft.class_1923;
import net.minecraft.class_1936;
import net.minecraft.class_1959;
import net.minecraft.class_1972;
import net.minecraft.class_2246;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.class_2902;
import net.minecraft.class_3341;
import net.minecraft.class_3443;
import net.minecraft.class_3485;
import net.minecraft.class_3532;
import net.minecraft.class_3612;
import net.minecraft.util.math.*;
import java.util.Random;

public class VolcanoGenerator extends class_3443 {
	private SimpleRadialNoise radiusNoise;
	private SimpleRadialNoise vegetationNoise;
	private SimpleRadialNoise chamberNoise;
	private OpenSimplexNoise chamberOreNoise;

	private int height;
	private int radius;
	private int lavaHeight;
	private int lavaTubeLength;
	private int baseY;
	private int chamberHeight;
	private boolean underwater;
	private long chamberOreSeed;

	private int centerX;
	private int centerZ;

	VolcanoGenerator(Random random, int centerX, int centerZ, class_1959 biome) {
		super(TerrestriaFeatures.VOLCANO_PIECE, 0);
		this.method_14926(null);

		this.centerX = centerX;
		this.centerZ = centerZ;

		radiusNoise = new SimpleRadialNoise(16, random.nextLong(), 0.75, 0.5);
		vegetationNoise = new SimpleRadialNoise(16, random.nextLong(), 0.25, 0.5);
		chamberNoise = new SimpleRadialNoise(16, random.nextLong(), 0.75, 0.5);
		chamberOreSeed = random.nextLong();
		chamberOreNoise = new OpenSimplexNoise(chamberOreSeed);

		if (biome == class_1972.field_9446 || biome == class_1972.field_9470 || biome == class_1972.field_9439 || biome == class_1972.field_9418 || biome == class_1972.field_9448) {
			height = 20 + random.nextInt(20);
			baseY = 30;
		} else if (biome == TerrestriaBiomes.VOLCANIC_ISLAND_SHORE) {
			height = 48 + random.nextInt(32);
			baseY = 45;
		} else {
			height = 32 + random.nextInt(64);
			baseY = 60;
		}

		if (height < 48) {
			radius = random.nextInt(height / 2) + height * 2;
		} else if (biome == TerrestriaBiomes.VOLCANIC_ISLAND_SHORE) {
			radius = random.nextInt(height / 3) + height / 4;
		} else {
			radius = random.nextInt(height * 3 / 4) + height / 2;
		}

		lavaHeight = (int) (height * shape(0.2));
		lavaTubeLength = Math.min(22, baseY - 20);

		// Make sure that the chamber doesn't go too deep
		int potentialChamberHeight = random.nextInt(Math.max(baseY - lavaTubeLength - 10, 0) + 1);
		chamberHeight = Math.max(potentialChamberHeight, 10);

		underwater = baseY + lavaHeight < 64;

		int radiusBound = class_3532.method_15384(radius * 1.5);

		this.field_15315 = new class_3341(centerX - radiusBound, 1, centerZ - radiusBound, centerX + radiusBound, 62 + height, centerZ + radiusBound);
	}

	public VolcanoGenerator(class_3485 manager, class_2487 tag) {
		super(TerrestriaFeatures.VOLCANO_PIECE, tag);

		radiusNoise = new SimpleRadialNoise(16, tag.method_10537("VRN"), 0.75, 0.5);
		vegetationNoise = new SimpleRadialNoise(16, tag.method_10537("VVN"), 0.25, 0.5);
		chamberNoise = new SimpleRadialNoise(16, tag.method_10537("VCN"), 0.75, 0.5);
		chamberOreSeed = tag.method_10537("VCON");
		chamberOreNoise = new OpenSimplexNoise(chamberOreSeed);

		height = tag.method_10550("VH");
		radius = tag.method_10550("VR");
		lavaHeight = tag.method_10550("VL");
		lavaTubeLength = tag.method_10550("VLT");
		baseY = tag.method_10550("Y");
		chamberHeight = tag.method_10550("VCH");
		underwater = tag.method_10577("VU");

		centerX = tag.method_10550("CX");
		centerZ = tag.method_10550("CZ");
	}

	private static double positionToAngle(double dist, double dX, double dZ) {
		// 0.0 to 0.5
		double angle = 0.5 * Math.asin(dX / dist) / Math.PI + 0.25;

		if (dZ < 0) {
			angle = 1.0 - angle;
		}

		return angle;
	}

	// Models the caldera shape of the volcano.
	private static double shape(double scaled) {
		// must be at least 0
		scaled = Math.max(scaled, 0.0);

		double curve = curve(1.0 - scaled);

		if (scaled <= 0.3) {
			curve -= (0.3 - scaled) * 2;
		}

		return curve;
	}

	// Models an S-curve.
	private static double curve(double progress) {
		// can't be greater than 1
		progress = Math.min(progress, 1.0);

		if (progress < 0.1) {
			return 2 * (progress - 0.1);
		} else if (progress <= 0.5) {
			return 2 * progress * progress;
		} else {
			double progressUntil = 1.0 - progress;

			return 1 - 2 * progressUntil * progressUntil;
		}
	}

	@Override
	protected void method_14943(class_2487 tag) {
		tag.method_10544("VRN", radiusNoise.getSeed());
		tag.method_10544("VVN", vegetationNoise.getSeed());
		tag.method_10544("VCN", chamberNoise.getSeed());
		tag.method_10544("VCON", chamberOreSeed);
		tag.method_10569("VH", height);
		tag.method_10569("VR", radius);
		tag.method_10569("VL", lavaHeight);
		tag.method_10569("VLT", lavaTubeLength);
		tag.method_10569("Y", baseY);
		tag.method_10569("VCH", chamberHeight);
		tag.method_10556("VU", underwater);

		tag.method_10569("CX", centerX);
		tag.method_10569("CZ", centerZ);
	}

	@Override
	public boolean method_14931(class_1936 world, Random random, class_3341 boundingBox, class_1923 chunkPos) {
		if (boundingBox.field_14377 < this.field_15315.field_14377 || boundingBox.field_14380 > this.field_15315.field_14380) {
			throw new IllegalArgumentException("Unexpected bounding box Y range in " + boundingBox + ", the Y range is smaller than the one we expected");
		}

		int chamberMiddle = baseY - lavaTubeLength - chamberHeight / 2;

		class_2338.class_2339 pos = new class_2338.class_2339();

		for (int z = boundingBox.field_14379; z <= boundingBox.field_14376; z++) {
			for (int x = boundingBox.field_14381; x <= boundingBox.field_14378; x++) {
				int dX = x - centerX;
				int dZ = z - centerZ;

				double dist = Math.sqrt(dZ * dZ + dX * dX);
				double angle = positionToAngle(dist, dX, dZ);

				// Fill in the lava chamber part!
				double chamberNoise = radiusNoise.sample(angle);
				double chamberScaled = (dist / radius) * chamberNoise;

				// chamberScaled = chamberScaled ^ 4
				chamberScaled *= chamberScaled;
				chamberScaled *= chamberScaled;

				double chamberShape = 1.0 - 16.0 * chamberScaled;
				int chamberDY = (int) (chamberHeight * 0.5 * chamberShape);

				if (chamberShape > 0.0) {
					for (int dY = -chamberDY; dY <= chamberDY; dY++) {
						pos.method_10103(x, chamberMiddle + dY, z);
						world.method_8652(pos, class_2246.field_10164.method_9564(), 2);
						world.method_8405().method_8676(pos, world.method_8316(pos).method_15772(), 0);
					}
				} else if (chamberShape > -0.1) {
					pos.method_10103(x, chamberMiddle, z);
					world.method_8652(pos, pickRandomChamberBlock(true, dX, dZ), 2);
				}

				// The center of the volcano is a lava tube, arranged in a plus sign shape.

				if (dZ == 0 && (dX >= -1 && dX <= 1) || dX == 0 && (dZ >= -1 && dZ <= 1)) {
					for (int dY = -(lavaTubeLength + chamberHeight); dY < lavaHeight; dY++) {
						pos.method_10103(x, baseY + dY, z);

						if (underwater && dY == lavaHeight - 1) {
							class_2680 state = random.nextInt(4) == 0 ? class_2246.field_10092.method_9564() : class_2246.field_10540.method_9564();

							world.method_8652(pos, state, 2);
						} else {
							world.method_8652(pos, class_2246.field_10164.method_9564(), 2);
							world.method_8405().method_8676(pos, world.method_8316(pos).method_15772(), 0);
						}
					}

					continue;
				} else if (chamberShape > 0.0) {
					world.method_8652(pos.method_10103(x, chamberMiddle + chamberDY + 1, z), pickRandomChamberBlock(true, dX, dZ), 2);
					world.method_8652(pos.method_10103(x, chamberMiddle - chamberDY - 1, z), pickRandomChamberBlock(false, dX, dZ), 2);
				}

				// Otherwise, proceed with normal generation. Sample the necessary values.

				double noise = radiusNoise.sample(angle);
				double vegetation = vegetationNoise.sample(angle) + random.nextDouble() * 0.15;

				double scaled = (dist / radius) * noise;
				int columnHeight = (int) (shape(scaled) * height);
				class_2680 top = TerrestriaBlocks.BASALT.plain.full.method_9564();

				// Below bedrock, skip.

				if (columnHeight + baseY <= 0) {
					continue;
				}

				// Add column height variance, for some random noise in the volcano shape.

				if (scaled > 0.2 && scaled < 0.35) {
					columnHeight += random.nextInt(2);
				} else if (scaled >= 0.35 && scaled <= 0.8 && random.nextInt(4) == 0) {
					columnHeight += 1;
				}

				// Set the vegetation / surface block to be used if the conditions are right.

				if (scaled > 0.3) {
					double scaledHeight = (double) (columnHeight) / (double) (lavaHeight);
					if (scaledHeight < vegetation) {
						if (columnHeight < 4) {
							top = TerrestriaBlocks.BASALT_SAND.method_9564();
						} else {
							top = TerrestriaBlocks.BASALT_GRASS_BLOCK.method_9564();
						}
					}
				}

				// Place the basalt column to the specified height.

				int startY = world.method_8589(class_2902.class_2903.field_13195, x, z) - baseY;

				for (int dY = startY; dY < columnHeight - 1; dY++) {
					pos.method_10103(x, baseY + dY, z);

					if (world.method_8320(pos).method_11588() || world.method_8316(pos).method_15772() == class_3612.field_15910) {
						world.method_8652(pos, TerrestriaBlocks.BASALT.plain.full.method_9564(), 2);
					}
				}

				// Surround the lava tube with basalt

				if ((Math.abs(dX) == 1 && Math.abs(dZ) == 1) || (Math.abs(dX) == 2 && dZ == 0) || (dX == 0 && Math.abs(dZ) == 2)) {
					startY = chamberMiddle + chamberDY + 1;
					int endY = baseY + columnHeight - 1;

					for (int y = startY; y < endY; y++) {
						pos.method_10103(x, y, z);

						world.method_8652(pos, TerrestriaBlocks.BASALT.plain.full.method_9564(), 2);
					}
				}

				// Some complex top block logic:
				// If the volcano is within the ocean, revert the top block to basalt. However, underwater volcanoes
				// have some magma on the sides, this is the one override.
				// Otherwise, if around the top areas of the volcano, place a few lava blocks to flow down the mountain.

				pos.method_10103(x, baseY + columnHeight, z);
				boolean lava = false;

				if (baseY < 60 || !world.method_8320(pos).method_11588()) {
					if (underwater && random.nextInt(80) == 0) {
						top = class_2246.field_10092.method_9564();
					} else {
						top = TerrestriaBlocks.BASALT.plain.full.method_9564();
					}
				} else if (scaled > 0.25 && scaled < 0.35) {
					if (!underwater && random.nextInt(320) == 0) {
						top = class_2246.field_10164.method_9564();
						lava = true;
					}
				}

				// Finally, actually set the top block. If lava, it needs a fluid tick update to get going.

				pos.method_10098(class_2350.field_11033);

				if ((world.method_8320(pos).method_11588() || world.method_8316(pos).method_15772() == class_3612.field_15910) && startY < columnHeight) {
					world.method_8652(pos, top, 2);

					if (lava) {
						world.method_8405().method_8676(pos, world.method_8316(pos).method_15772(), 0);
					}
				}

				// If within the top caldera, set lava / magma / obsidian based on whether this is an underwater volcano or not.
				// Effectively, fill the caldera with lava up to the lavaHeight height.

				if (scaled <= 0.3) {
					for (int dY = columnHeight; dY < lavaHeight; dY++) {
						pos.method_10103(x, baseY + dY, z);

						if (underwater && dY == lavaHeight - 1) {
							class_2680 state = random.nextInt(6) == 0 ? class_2246.field_10092.method_9564() : class_2246.field_10540.method_9564();

							world.method_8652(pos, state, 2);
						} else {
							world.method_8652(pos, class_2246.field_10164.method_9564(), 2);
						}
					}
				}
			}
		}

		return true;
	}

	private class_2680 pickRandomChamberBlock(boolean top, int dX, int dZ) {
		if (!top) {
			dX = -dX;
		}

		double goldNoise = chamberOreNoise.sample(dX * 0.05, dZ * 0.05);
		double diamondNoise = chamberOreNoise.sample(-dX * 0.2, dZ * 0.2);
		double obsidianNoise = chamberOreNoise.sample(dX * 0.05, -dZ * 0.05);

		if (diamondNoise > 0.70) {
			return class_2246.field_10442.method_9564();
		} else if (goldNoise < -0.75) {
			return class_2246.field_10571.method_9564();
		} else {
			return obsidianNoise > 0.25 ? class_2246.field_10540.method_9564() : TerrestriaBlocks.BASALT.plain.full.method_9564();
		}
	}
}
